@@ -5,6 +5,12 @@
 import hashlib, re, string, struct
 from . import Util
 
+try:
+	Image = None
+	from PIL import Image
+except:
+	pass
+
 class RomFileDMG:
 	ROMFILE_PATH = None
 	ROMFILE = bytearray()
@@ -51,6 +57,33 @@ class RomFileDMG:
 		self.CalcChecksumGlobal(True)
 		return self.ROMFILE[0:0x200]
 	
+	def LogoToImage(self, data, valid=True):
+		if Image is None: return False
+		img = Image.new(mode='P', size=(48, 8))
+		if valid:
+			img.putpalette([ 255, 255, 255, 0, 0, 0 ])
+		else:
+			img.putpalette([ 255, 255, 255, 255, 0, 0 ])
+		img.info["transparency"] = 0
+		pixels = img.load()
+
+		for y in range(0, 8):
+			i = int((y/2)%2) + int(y/4)*24
+			x = 0
+			ix = 0
+			while True:
+				x += 1
+				nibble = (data[i]&0xF) if (y%2) else (data[i]>>4)
+				for b in range(3, -1, -1):
+					if ((nibble>>b)&1):
+						pixels[ix, y] = 1
+					else:
+						pixels[ix, y] = 0
+					ix += 1
+				i += 2
+				if x >= 12: break
+		return img
+
 	def GetHeader(self, unchanged=False):
 		buffer = self.ROMFILE
 		data = {}
@@ -61,6 +94,8 @@ class RomFileDMG:
 		data["cgb"] = int(buffer[0x143])
 		data["sgb"] = int(buffer[0x146])
 		data["old_lic"] = int(buffer[0x14B])
+		temp = self.LogoToImage(buffer[0x104:0x134], data["logo_correct"])
+		if temp is not False and not data["empty"]: data["logo"] = temp
 		
 		if data["cgb"] in (0x80, 0xC0):
 			data["game_title_raw"] = bytearray(buffer[0x134:0x143]).decode("ascii", "replace")
@@ -70,6 +105,12 @@ class RomFileDMG:
 		game_title = re.sub(r"(\x00+)$", "", game_title)
 		game_title = re.sub(r"((_)_+|(\x00)\x00+|(\s)\s+)", "\\2\\3\\4", game_title).replace("\x00", "_")
 		game_title = ''.join(filter(lambda x: x in set(string.printable), game_title))
+		data["game_code"] = ""
+		if data["cgb"] in (0x80, 0xC0):
+			if len(data["game_title_raw"].rstrip("\x00")) == 15:
+				if data["game_title_raw"][-4:][0] in ("A", "B", "H", "K", "V") and data["game_title_raw"][-4:][3] in ("A", "B", "D", "E", "F", "I", "J", "K", "P", "S", "U", "X", "Y"):
+					data["game_code"] = game_title[-4:]
+					game_title = game_title[:-4].rstrip("_")
 		data["game_title"] = game_title
 
 		data["maker_code"] = format(int(buffer[0x14B]), "02X")
@@ -198,32 +239,38 @@ class RomFileDMG:
 
 			# Unlicensed Sachen MMC1/MMC2
 			elif len(buffer) >= 0x280:
+				sachen_version = 0
 				sachen_hash = hashlib.sha1(buffer[0x200:0x280]).digest()
 				if sachen_hash == bytearray([ 0x73, 0x9B, 0x46, 0x86, 0x97, 0x1C, 0x0B, 0xAE, 0xAF, 0x26, 0xF1, 0x73, 0xAC, 0xAE, 0x4B, 0x2B, 0xBF, 0x00, 0x70, 0x77 ]):
 					data["rom_size_raw"] = 0x03
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x8793
 					data["game_title"] = "SACHEN 4B-001"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0x22, 0x8B, 0x40, 0x4F, 0x1C, 0xCF, 0x4D, 0xDC, 0x4D, 0xF2, 0x35, 0xF3, 0x7B, 0x6D, 0x61, 0x5E, 0xBE, 0xF1, 0xEF, 0x42 ]):
 					data["rom_size_raw"] = 0x01
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0xB180
 					data["game_title"] = "SACHEN 4B-002"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0xF9, 0xB8, 0x6A, 0x8F, 0x2E, 0x8B, 0x31, 0xD4, 0xC5, 0x02, 0xC8, 0x80, 0x75, 0x35, 0x9C, 0x02, 0xB3, 0xB5, 0x68, 0x01 ]):
 					data["rom_size_raw"] = 0x03
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x7CFD
 					data["game_title"] = "SACHEN 4B-004"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0xD6, 0xB5, 0x33, 0x81, 0x1A, 0x01, 0x0D, 0x4D, 0x1C, 0xCC, 0x5A, 0x2C, 0x34, 0x9D, 0x0F, 0x63, 0xD3, 0xF4, 0x9D, 0x34 ]):
 					data["rom_size_raw"] = 0x03
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x80DF
 					data["game_title"] = "SACHEN 4B-005"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0x3E, 0x56, 0xCC, 0x2D, 0xDF, 0xE0, 0x00, 0xED, 0x53, 0xA7, 0x9D, 0x62, 0xC8, 0xBF, 0x7F, 0x20, 0x27, 0x47, 0xCD, 0x8E ]):
 					data["rom_size_raw"] = 0x03
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x55E6
 					data["game_title"] = "SACHEN 4B-006"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0xD8, 0x24, 0xD2, 0xB2, 0x71, 0x6B, 0x08, 0xFA, 0xEA, 0xA4, 0xFB, 0xD9, 0x7D, 0x81, 0x94, 0x57, 0x46, 0x77, 0x91, 0x60 ]):
 					sachen_hash2 = hashlib.sha1(buffer[0x0:0x80]).digest()
 					if sachen_hash2 == bytearray([ 0x20, 0x17, 0x3B, 0x95, 0xFB, 0x1A, 0xAC, 0x81, 0x6C, 0x66, 0xF5, 0x62, 0x11, 0xF4, 0x96, 0xFB, 0xFB, 0x88, 0x03, 0xBC ]):
@@ -231,91 +278,114 @@ class RomFileDMG:
 						data["mapper_raw"] = 0x204
 						data["rom_checksum"] = 0x657A
 						data["game_title"] = "SACHEN 8B-001"
+						sachen_version = 2
 					else:
 						data["rom_size_raw"] = 0x03
 						data["mapper_raw"] = 0x204
 						data["rom_checksum"] = 0x8E9F
 						data["game_title"] = "SACHEN 4B-007"
+						sachen_version = 1
 				elif sachen_hash == bytearray([ 0x19, 0x3E, 0xF8, 0xE2, 0x12, 0x8A, 0x24, 0x10, 0xFE, 0xE9, 0xEA, 0x27, 0xC9, 0x1B, 0xC4, 0xDD, 0x04, 0x74, 0x1B, 0xA8 ]):
 					data["rom_size_raw"] = 0x03
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x99C7
 					data["game_title"] = "SACHEN 4B-008"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0xA5, 0x07, 0xCB, 0xB0, 0x63, 0x7A, 0xE7, 0x1A, 0xF2, 0xC8, 0x32, 0x9B, 0xA6, 0x6D, 0xC4, 0x21, 0x68, 0x78, 0xE5, 0x39 ]):
 					data["rom_size_raw"] = 0x03
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0xCDD3
 					data["game_title"] = "SACHEN 4B-009"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0x18, 0xEC, 0x2B, 0x15, 0x97, 0xD7, 0x80, 0x51, 0x58, 0xB2, 0xB8, 0x53, 0xA7, 0x00, 0xD7, 0x0B, 0xCE, 0x0A, 0xB3, 0xFF ]):
 					data["rom_size_raw"] = 0x01
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x3889
 					data["game_title"] = "SACHEN 4B-003"
+					sachen_version = 1
 				elif sachen_hash == bytearray([ 0x96, 0x1C, 0xE3, 0x5D, 0x3A, 0x81, 0x44, 0x95, 0xCF, 0x42, 0x92, 0x42, 0x30, 0x83, 0x14, 0x17, 0xA9, 0xBF, 0xE0, 0x9F ]):
 					data["rom_size_raw"] = 0x06
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x3934
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 31B-001"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0xB8, 0x59, 0x61, 0x1C, 0x03, 0xAF, 0x5F, 0x7F, 0x50, 0x3E, 0x8C, 0xB0, 0x9C, 0x81, 0x4A, 0x0C, 0xE8, 0xBA, 0xB5, 0x99 ]):
 					data["rom_size_raw"] = 0x06
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x2E45
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 31B-001"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0xF5, 0xC6, 0xC2, 0xE6, 0xA6, 0xF2, 0xEE, 0x86, 0x29, 0x22, 0x3D, 0x7C, 0x72, 0xF9, 0xDD, 0x6F, 0x32, 0x0A, 0xA0, 0x9D ]):
 					data["rom_size_raw"] = 0x04
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x125B
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 8B-001"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0xF2, 0x8A, 0xDF, 0x84, 0xBA, 0x56, 0x8C, 0x54, 0xF9, 0x4B, 0x25, 0xFA, 0x12, 0x92, 0x4E, 0xD6, 0x7D, 0xD1, 0x7E, 0x9D ]):
 					data["rom_size_raw"] = 0x04
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x598F
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 8B-002"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0x1C, 0x08, 0x6F, 0x94, 0xD8, 0xFD, 0x40, 0x4D, 0xA3, 0x85, 0xCE, 0x57, 0x35, 0xF3, 0x43, 0x92, 0xEE, 0xB7, 0x26, 0xE1 ]):
 					data["rom_size_raw"] = 0x04
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x6485
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 8B-003"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0x2C, 0xFD, 0xE1, 0x8D, 0x2C, 0x57, 0xBA, 0xDB, 0xC0, 0xF8, 0xDF, 0x52, 0x79, 0x38, 0x44, 0x56, 0x3B, 0xB0, 0xA0, 0xDE ]):
 					data["rom_size_raw"] = 0x04
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x02A4
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 8B-004"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0x4E, 0xEA, 0x3C, 0x0A, 0x23, 0x5C, 0xF9, 0x2D, 0xC6, 0x22, 0xC2, 0x21, 0xD3, 0xBB, 0x73, 0x3B, 0xA7, 0x21, 0xFB, 0x78 ]):
 					data["rom_size_raw"] = 0x02
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0xA709
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0x3F, 0xE9, 0xB7, 0xAB, 0xBC, 0x18, 0x95, 0x60, 0x80, 0xF7, 0xDF, 0x9B, 0x5E, 0x5A, 0x0C, 0x9F, 0x18, 0x63, 0x34, 0x7B ]):
 					data["rom_size_raw"] = 0x04
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x929D
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 1B-003"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0xD8, 0x9B, 0x0D, 0x55, 0x48, 0x97, 0x7F, 0xD5, 0x0E, 0x46, 0x20, 0xD6, 0x9E, 0x0B, 0x8C, 0x6B, 0x05, 0xD4, 0x8F, 0x2C ]):
 					data["rom_size_raw"] = 0x04
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x2F50
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN 4B-003"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0x8B, 0x98, 0xB1, 0xD3, 0x6B, 0x84, 0x66, 0x51, 0xC0, 0x23, 0x19, 0xF2, 0xDC, 0xD3, 0xF4, 0x97, 0xDB, 0x39, 0x47, 0xE7 ]):
 					data["rom_size_raw"] = 0x06
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x9769
 					data["cgb"] = 0x80
 					data["game_title"] = "SACHEN"
+					sachen_version = 2
 				elif sachen_hash == bytearray([ 0xD0, 0xAE, 0xC9, 0xFB, 0xF0, 0x8D, 0x7A, 0x72, 0x34, 0x8E, 0x96, 0xB6, 0x75, 0x6B, 0x30, 0xC1, 0xCB, 0xF6, 0x2F, 0x00 ]):
 					data["rom_size_raw"] = 0x06
 					data["mapper_raw"] = 0x204
 					data["rom_checksum"] = 0x0346
 					data["game_title"] = "SACHEN 6B-001"
+					sachen_version = 2
+
+				if sachen_version == 1:
+					temp = self.LogoToImage(buffer[0x184:0x184+0x30])
+					if temp is not False: data["logo_sachen"] = temp
+				elif sachen_version == 2:
+					temp = self.LogoToImage(buffer[0x104:0x104+0x30])
+					if temp is not False: data["logo_sachen"] = temp
 
 		if data["mapper_raw"] in Util.DMG_Header_Mapper:
 			data["mapper"] = Util.DMG_Header_Mapper[data["mapper_raw"]]
